/*****************************************************************************
 *
 * FILE:	cookie.c
 * DESCRIPTION:	Cookie manipulator
 * DATE:	Sat, Mar 16 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: cookie.c,v 1.3 2002/06/27 06:55:24 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if	USE_WKF
#include <wkf.h>
#endif	/* USE_WKF */
#include "cockatrice.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	COOKIE_TABLE_SIZE	11
typedef struct _cookie {	/* cookie structure (cf. rfc2109) */
  String		name;		/* NAME */
  String		value;		/* VALUE */
  String		comment;	/* $Comment */
  String		domain;		/* $Domain */
  unsigned int		max_age;	/* $Max-Age */
  String		path;		/* $Path */
  bool_t		secure;		/* $Secure */
  unsigned int		version;	/* $Version=1 */
  struct _cookie *	next;
} Cookie;


/*****************************************************************************
 *
 *	definition of local functions
 *
 *****************************************************************************/
static void	cleanup(Cookie *);
static int	parseString(register String, String *, String *);
static int	putCookieParameter(const String, const String, unsigned int,
				   const String, const String, unsigned int,
				   const String, bool_t);
static Cookie *	lookup(const String);


/*****************************************************************************
 *
 *	definition of local variables
 *
 *****************************************************************************/
static Cookie *	cookie[COOKIE_TABLE_SIZE];	/* for Cookie field */
static Cookie *	setCookie[COOKIE_TABLE_SIZE];	/* for Set-Cookie field */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	cookieInit
 * DESCRIPTION:		initialize a cookie table
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
cookieInit(void)
{
  unsigned int	i;

  for (i = 0; i < COOKIE_TABLE_SIZE; i++) {
    cookie[i] = setCookie[i] = NULL;
  }
}

/*
 * FUNCTION NAME:	cookieRelease
 * DESCRIPTION:		release memory area of cookie table
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
cookieRelease(void)
{
  unsigned int	i;

  for (i = 0; i < COOKIE_TABLE_SIZE; i++) {
    if (cookie[i] != NULL) {
      cleanup(cookie[i]);
      cookie[i] = NULL;
    }
    if (setCookie[i] != NULL) {
      cleanup(setCookie[i]);
      setCookie[i] = NULL;
    }
  }
}

static void
cleanup(c)
	Cookie *	c;
{
  register Cookie *	cp;
  register Cookie *	tmp;

  for (cp = c; cp != NULL; cp = tmp) {
    tmp = cp->next;
    if (cp->name != NULL) {
      free((String)cp->name);
    }
    if (cp->value != NULL) {
      free((String)cp->value);
    }
    if (cp->comment != NULL) {
      free((String)cp->comment);
    }
    if (cp->domain != NULL) {
      free((String)cp->domain);
    }
    cp->max_age = 0;
    if (cp->path != NULL) {
      free((String)cp->path);
    }
    cp->secure	= FALSE;
    cp->version	= 1;
    free((Cookie *)cp);
  }
}

/*
 * FUNCTION NAME:	cookieParse
 * DESCRIPTION:		parse Cookie string
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
cookieParse(void)
{
  String	string	= getenv("HTTP_COOKIE");
  String	name	= NULL;	/* NAME */
  String	value	= NULL;	/* VALUE */
  unsigned int	version	= 0;	/* $Version */
  String	domain	= NULL;	/* $Domain */
  String	path	= NULL;	/* $Path */
  unsigned int	max_age	= 0;	/* $Max-Age */
  String	comment	= NULL;	/* $Comment */
  bool_t	secure	= FALSE;/* $Secure */

  if (string == NULL) {
    return -1;	/* No cookies? or error? */
  }

#if	DEBUG
  fprintf(debugLog, "Http-Cookie: %s\n", string);
#endif	/* DEBUG */

#if	USE_WKF
  wkfSetLineEndCode(LE_LF);
#endif	/* USE_WKF */

  while ((string = strtok(string, ";,")) != NULL) {
    String	Pname;	/* parameter name */
    String	Pvalue;	/* parameter value */

    /* skip white space */
    while (*string == SPC || *string == TAB) {
      string++;
    }

#if	DEBUG
    fprintf(debugLog, "String = %s\n", string);
#endif	/* DEBUG */
    if (parseString(string, &Pname, &Pvalue) == 0) {
      name  = unescapeURI(Pname);
      value = unescapeURI(Pvalue);

#if	DEBUG
      fprintf(debugLog, "Parse: %s = %s\n", Pname, Pvalue);
#endif	/* DEBUG */
      /* reserved parameter */
      if (Pname[0] == '$') {
	if (strcasecmp(Pname, "$Version") == 0) {
	  version = (unsigned int)atoi(Pvalue);
	}
	else if (strcasecmp(Pname, "$Domain") == 0) {
	  domain = Pvalue;
	}
	else if (strcasecmp(Pname, "$Path") == 0) {
	  path = Pvalue;
	}
	else if (strcasecmp(Pname, "$Max-Age") == 0) {
	  max_age = (unsigned int)atoi(Pvalue);
	}
	else if (strcasecmp(Pname, "$Comment") == 0) {
	  comment = Pvalue;
	}
	else if (strcasecmp(Pname, "$Secure") == 0) {
	  secure = TRUE;
	}
	else {
	  /* invalid reserved parameter */
	}
      }
      /* user defined parameter */
      else {
#if	USE_WKF
	String		Kname;
	String		Kvalue;
	unsigned int	len;

	len	= strlen(name) + 1;
	Kname	= (String)calloc(len, sizeof(char));
	if (Kname != NULL &&
	    wkfConvertKanjiCodeOfString(KC_UNKNOWN, name,
					KC_EUC, Kname, len) != CONV_ERR) {
	  name = Kname;
	}
	len	= strlen(value) + 1;
	Kvalue	= (String)calloc(len, sizeof(char));
	if (Kvalue != NULL &&
	    wkfConvertKanjiCodeOfString(KC_UNKNOWN, value,
					KC_EUC, Kvalue, len) != CONV_ERR) {
	  value = Kvalue;
	}
#endif	/* USE_WKF */
	putCookieParameter(name, value, version, path, domain,
			   max_age, comment, secure);
#if	USE_WKF
	if (Kname != NULL) {
	    free((String)Kname);
	}
	if (Kvalue != NULL) {
	    free((String)Kvalue);
	}
#endif	/* USE_WKF */
#if	DEBUG
	fprintf(debugLog, "Waiting: %s = %s\n", name, value);
#endif	/* DEBUG */
      }
    }
    string = NULL;	/* DON'T REMOVE! */
  }

  return 0;
}

/*
 * FUNCTION NAME:	parseString
 * DESCRIPTION:		parse NAME=VALUE string of cookie
 * ARGUMENTS:		string, name pointer, value pointer
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
parseString(string, name, value)
	register String	string;
	String *	name;
	String *	value;
{
  String	p;

  /* initialize */
  *name = *value = NULL;

  /* $Secure is special parameter without a value */
  if (strncasecmp(string, "$Secure", 7) == 0) {
    *name = strdup(string);
    return 0;
  }

  /* NAME */
  p = (String)strchr(string, '=');
  if (p != NULL) {
    *p++   = EOL;
    *name  = strdup(string);
    string = ++p;	/* put position on first quote(") mark */
  }
  /* VALUE */
  p = (String)strchr(string, '"');
  if (p != NULL) {
    *p	   = EOL;
    *value = strdup(string);
  }
#if	DEBUG
  fprintf(debugLog, "Name[%s] Value[%s]\n", *name, *value);
#endif	/* DEBUG */

  return (*name == NULL || *value == NULL ? -1 : 0);
}

/*
 * FUNCTION NAME:	putCookieParameter
 * DESCRIPTION:		store 'NAME=VALUE' of cookie into the table
 * ARGUMENTS:		name, value, version, path, domain,
 *			max_age, comment, secure
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
putCookieParameter(name, value, version, path, domain, max_age, comment, secure)
	const String	name;
	const String	value;
	unsigned int	version;
	const String	path;
	const String	domain;
	unsigned int	max_age;
	const String	comment;
	bool_t		secure;
{
  Cookie *	cp;	/* new parameter pointer */
  unsigned int	h;	/* hash value */

  /* reject no name / no value */
  if (name == NULL || value == NULL) {
    return -1;
  }

  cp = (Cookie *)calloc(1, sizeof(Cookie));
  if (cp == NULL) {
    return -1;
  }

  h		= hash(name, COOKIE_TABLE_SIZE);
  cp->name	= strdup(name);
  cp->value	= strdup(value);
  cp->comment	= comment != NULL ? strdup(comment) : NULL;
  cp->domain	= domain != NULL ? strdup(domain) : NULL;
  cp->max_age	= max_age;
  cp->path	= path != NULL ? strdup(path) : NULL;
  cp->secure	= secure;
  cp->version	= version;
  cp->next	= cookie[h];
  cookie[h]	= cp;
#if	DEBUG
  fprintf(debugLog, "Registered: %s = %s\n", cp->name, cp->value);
#endif	/* DEBUG */

  return 0;
}

/*
 * FUNCTION NAME:	cookieGet
 * DESCRIPTION:		get Cookie value from the table
 * ARGUMENTS:		name
 * RETURN VALUE:	value
 */
String
cookieGet(name)
	const String	name;
{
  register Cookie *	cp;	/* working pointer */
  unsigned int		h;	/* hash value */

  if (name != NULL) {
    h = hash(name, COOKIE_TABLE_SIZE);
    for (cp = cookie[h]; cp != NULL; cp = cp->next) {
      if (cp->name != NULL && strcmp(name, cp->name) == 0) {
#if	DEBUG
	fprintf(debugLog, "Hit: %s = %s\n", cp->name, cp->value);
#endif	/* DEBUG */
	return cp->value;
      }
    }
  }

  return NULL;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	cookieSetName
 * DESCRIPTION:		set Cookie Name and Value
 * ARGUMENTS:		name, value
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
cookieSetName(name, value)
	const String	name;
	const String	value;
{
  Cookie *	cp;	/* new parameter pointer */
  unsigned int	h;	/* hash value */

  /* reject no name / no value */
  if (name == NULL || value == NULL) {
    return -1;
  }
  cp = (Cookie *)calloc(1, sizeof(Cookie));
  if (cp == NULL) {
    return -1;
  }

  h		= hash(name, COOKIE_TABLE_SIZE);
  cp->name	= strdup(name);
  cp->value	= strdup(value);
  cp->comment	= NULL;
  cp->domain	= NULL;
  cp->max_age	= 0;
  cp->path	= NULL;
  cp->secure	= FALSE;
  cp->version	= 1;
  cp->next	= setCookie[h];
  setCookie[h]	= cp;

#if	DEBUG
  fprintf(debugLog, "Cookie ... set (%s = %s)\n", name, value);
#endif	/* DEBUG */
  return 0;
}

/*
 * FUNCTION NAME:	cookieSetComment
 * DESCRIPTION:		set Cookie Comment
 * ARGUMENTS:		name, comment
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
cookieSetComment(name, comment)
	const String	name;
	const String	comment;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL && comment != NULL) {
    cp->comment = strdup(comment);
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cookieSetDomain
 * DESCRIPTION:		set Cookie Domain
 * ARGUMENTS:		name, domain
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
cookieSetDomain(name, domain)
	const String	name;
	const String	domain;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL && domain != NULL) {
    cp->domain = strdup(domain);
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cookieSetMaxAge
 * DESCRIPTION:		set Cookie MaxAge
 * ARGUMENTS:		name, max-age
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
cookieSetMaxAge(name, max_age)
	const String	name;
	unsigned int	max_age;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL) {
    cp->max_age = max_age;
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cookieSetPath
 * DESCRIPTION:		set Cookie Path
 * ARGUMENTS:		name, path
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
cookieSetPath(name, path)
	const String	name;
	const String	path;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL && path != NULL) {
    cp->path = strdup(path);
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cookieSetSecure
 * DESCRIPTION:		set Cookie Secure
 * ARGUMENTS:		name, secure
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
cookieSetSecure(name, secure)
	const String	name;
	bool_t		secure;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL) {
    cp->secure = secure;
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	lookup
 * DESCRIPTION:		look up a SetCookie data for Cookie name
 * ARGUMENTS:		name
 * RETURN VALUE:	Successful: Cookie pointer, otherwise: NULL
 */
static Cookie *
lookup(name)
	const String	name;
{
  Cookie *	cp;	/* working pointer */
  unsigned int	h;	/* hash value */

  if (name != NULL) {
    h = hash(name, COOKIE_TABLE_SIZE);
    for (cp = setCookie[h]; cp != NULL; cp = cp->next) {
      if (cp->name != NULL && strcmp(name, cp->name) == 0) {
	return cp;
      }
    }
  }

  return NULL;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	sendResponseHeader
 * DESCRIPTION:		send HTTP response header field
 * ARGUMENTS:		Content-Type
 * RETURN VALUE:	none
 */
void
sendResponseHeader(ctype)
	const String	ctype;
{
  unsigned int	i;

  /* Content-Type response field */
  fputs("Content-Type: ", stdout);
  if (ctype != NULL) {
    fprintf(stdout, "%s" CRLF, ctype);
  }
  else {
    fputs("text/html" CRLF, stdout);
  }

  /* Set-Cookie response field */
  for (i = 0; i < COOKIE_TABLE_SIZE; i++) {
    register Cookie *	cp;

    if (setCookie[i] == NULL) {
      continue;
    }
    for (cp = setCookie[i]; cp != NULL; cp = cp->next) {
      fputs("Set-Cookie: ", stdout);
      fprintf(stdout, "%s=\"%s\"", escapeURI(cp->name), escapeURI(cp->value));
      if (cp->comment != NULL) {
	fprintf(stdout, ";Comment=\"%s\"", cp->comment);
      }
      if (cp->domain != NULL) {
	fprintf(stdout, ";Domain=\"%s\"", cp->domain);
      }
      if (cp->max_age > 0) {
	fprintf(stdout, ";Max-Age=\"%d\"", cp->max_age);
      }
      if (cp->path != NULL) {
	fprintf(stdout, ";Path=\"%s\"", cp->path);
      }
      if (cp->secure) {
	fputs(";Secure", stdout);
      }
      fprintf(stdout, ";Version=\"%d\"" CRLF, cp->version);
#if	DEBUG
      fputs("Set-Cookie: ", debugLog);
      fprintf(debugLog, "%s=\"%s\"", escapeURI(cp->name), escapeURI(cp->value));
      if (cp->comment != NULL) {
	fprintf(debugLog, ";Comment=\"%s\"", cp->comment);
      }
      if (cp->domain != NULL) {
	fprintf(debugLog, ";Domain=\"%s\"", cp->domain);
      }
      if (cp->max_age > 0) {
	fprintf(debugLog, ";Max-Age=\"%d\"", cp->max_age);
      }
      if (cp->path != NULL) {
	fprintf(debugLog, ";Path=\"%s\"", cp->path);
      }
      if (cp->secure) {
	fputs(";Secure", debugLog);
      }
      fprintf(debugLog, ";Version=\"%d\"" CRLF, cp->version);
#endif	/* DEBUG */
    }
  }
  fputs(CRLF, stdout);
}

