/*****************************************************************************
 *
 * FILE:	lock.c
 * DESCRIPTION:	lock utilities for simple CGI library
 * DATE:	Sat, Mar 16 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 2002 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: lock.c,v 1.3 2002/06/15 03:32:33 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include "cockatrice.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
typedef struct _LockManager {
  String		name;
  String		path;
  struct _LockManager *	next;
} LockManager;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static String		basename(const String);
#if	0
static struct flock *	fileLock(short, short);
#endif


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static LockManager *	locks = NULL;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	lock
 * DESCRIPTION:		lock file exclusively
 * ARGUMENTS:		filename
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
lock(filename)
	String	filename;
{
  LockManager *	lmp;
  int		wait = 5;	/* max waiting time */

  if (filename == NULL) {
    cgi_errno = CGIERR_INVALID_ARGUMENTS;
    return -1;
  }

  lmp = (LockManager *)calloc(1, sizeof(LockManager));
  if (lmp == NULL) {
    cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
    return -1;
  }

  if (asprintf(&lmp->path, "%s.lock", basename(filename)) == -1) {
    cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
    return -1;
  }
  lmp->name = strdup(filename);

  /* lock the file */
  while (symlink(".", lmp->path) == -1) {
    if (wait-- < 0) {
      free((String)lmp->name);
      free((String)lmp->path);
      free((LockManager *)lmp);
      cgi_errno = CGIERR_OTHER_PROCESS_USING;
      return -1;
    }
#if	HAVE_USLEEP
    usleep(200000);	/* 0.2 second */
#else
    sleep(1);
#endif	/* HAVE_USLEEP */
  }
  lmp->next	= locks;
  locks		= lmp;

  return 0;
}

/*
 * FUNCTION NAME:	unlock
 * DESCRIPTION:		unlock file
 * ARGUMENTS:		filename
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
unlock(filename)
	const String	filename;
{
  register LockManager *	lmp;

  for (lmp = locks; lmp != NULL; lmp = lmp->next) {
    if (strcmp(lmp->name, filename) == 0) {
      if (unlink(lmp->path) != -1) {
	free((String)lmp->name); lmp->name = NULL;
	free((String)lmp->path); lmp->path = NULL;
	return 0;
      }
    }
  }

  return -1;
}

/*
 * FUNCTION NAME:	unlockAll
 * DESCRIPTION:		unlock all locked files
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
unlockAll(void)
{
  register LockManager *	lmp;
  register LockManager *	tmp;

  for (lmp = locks; lmp != NULL; lmp = tmp) {
    tmp = lmp->next;
    if (lmp->path != NULL) {
      unlink(lmp->path);
    }
    if (lmp->name != NULL) {
      free((String)lmp->name);
    }
    if (lmp->path != NULL) {
      free((String)lmp->path);
    }
    free((LockManager *)lmp);
  }
}

/*
 * Function Name:	basename
 * Description:		return filename portion of pathname
 * Arguments:		file path
 * Return Value:	Successful: basename, otherwise: NULL
 */
static String
basename(path)
	String	path;
{
  String	p;

  if (path == NULL) {
    return NULL;
  }

  p = (String)strrchr(path, '/');
  if (p == NULL) {
    return path;
  }
  return ++p;
}


/*****************************************************************************
 *****************************************************************************/

#if	0
/*
 * FUNCTION NAME:	lock
 * DESCRIPTION:		lock file exclusively
 * ARGUMENTS:		filename
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
lock(filename)
	const String	filename;
{
  int	fd;
  fcntl(fd, F_SETLK, fileLock(F_RDLCK, SEEK_SET));	/* shared */
  fcntl(fd, F_SETLK, fileLock(F_WRLCK, SEEK_SET));	/* exclusive */
  fcntl(fd, F_SETLK, fileLock(F_WRLCK, SEEK_END));	/* append */
}

/*
 * FUNCTION NAME:	unlock
 * DESCRIPTION:		unlock file
 * ARGUMENTS:		filename
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
unlock(filename)
	const String	filename;
{
  int	fd;
  return fcntl(fd, F_SETLK, fileLock(F_UNLCK, SEEK_SET));
}


static struct flock *
fileLock(type, whence)
	short	type;
	short	whence;
{
  static struct flock	ret;

  ret.l_start	= 0;
  ret.l_len	= 0;
  ret.l_pid	= getpid();
  ret.l_type	= type;
  ret.l_whence	= whence;

  return &ret;
}
#endif

