/*****************************************************************************
 *
 * FILE:	session.c
 * DESCRIPTION:	session manager for simple CGI library
 * DATE:	Tue, Mar 19 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 2002 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: session.c,v 1.12 2002/07/22 08:36:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <fcntl.h>
#include <sys/types.h>
#include <limits.h>
#include <db.h>
#include <sha.h>
#include <sys/time.h>
#include "cockatrice.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	DEF_SESSION_DB		"cockatrice.session.db"
#define	TMP_SESSION_DB		"/tmp/" ## DEF_SESSION_DB
#define	DEF_SESSION_EXPIRES	(20)	/* 20 min. */
#define	DELETE_WAITING		(30)	/* 30 min. */
#define	KEY_EXPIRES		"Cockatrice.Session.Expires"


/*****************************************************************************
 *
 *	definition of local functions
 *
 *****************************************************************************/
static int	expireSession(const DBT *, const DBT *);
static int	checkExpireSession(void);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static DB *		session = NULL;
static unsigned int	expires = (DEF_SESSION_EXPIRES * 60);
static unsigned int	waiting = ((DEF_SESSION_EXPIRES + DELETE_WAITING) * 60);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	sessionOpenDB
 * DESCRIPTION:		open session database
 * ARGUMENTS:		database name
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
sessionOpenDB(dbname)
	const String	dbname;
{
  unsigned int	minutes;

  sessionCloseDB();

  session = dbopen(dbname, O_CREAT|O_RDWR|O_EXLOCK, 0600, DB_HASH, NULL);
  if (session == NULL) {
    cgi_errno = CGIERR_DB_OPEN_FAILURE;
    return -1;
  }
#if	1
  minutes = sessionGetExpires();
  if (minutes == -1) {
    minutes = DEF_SESSION_EXPIRES;
  }
  expires = minutes * 60;
  waiting = (minutes + DELETE_WAITING) * 60;
#endif

  checkExpireSession();

  return 0;
}

/*
 * FUNCTION NAME:	sessionCloseDB
 * DESCRIPTION:		close session manager
 * ARGUMENTS:		none
 * RETURN VALUE:	same as db->close()
 */
int
sessionCloseDB(void)
{
  int	ret;

  if (session == NULL) {
    return 0;
  }

  ret	  = session->close(session);
  session = NULL;

  return ret;
}

/*
 * FUNCTION NAME:	sessionGenerateID0
 * DESCRIPTION:		generate a session id
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: ID, otherwise: NULL
 */
String
sessionGenerateID0(void)
{
  int			n;
  unsigned char *	text;

  n = asprintf((char **)&text, "%ld-%s(%u)%s:%u.%ld",
		(long)time(NULL), envScriptName(), getpid(),
		envRemoteAddr(), envRemotePort(), random());
  if (n != -1) {
    String	id;
    
    id = sessionGenerateID1(text);
    free((unsigned char *)text);

    return id;
  }

  return NULL;
}

/*
 * FUNCTION NAME:	sessionGenerateID
 * DESCRIPTION:		generate a session id
 * ARGUMENTS:		string
 * RETURN VALUE:	Successful: ID, otherwise: NULL
 */
String
sessionGenerateID1(text)
	const String	text;
{
  struct timeval	now;
  SHA_CTX		context;
  unsigned char		digest[SHA_DIGEST_LENGTH];
  unsigned char		id[SHA_DIGEST_LENGTH*2+1];
  DBT			key;
  DBT			val;
  int			i;

  if (gettimeofday(&now, NULL) == -1) {
    return NULL;
  }

  SHA1_Init(&context);
  SHA1_Update(&context, text, strlen(text));
  SHA1_Final(digest, &context);

  memset(id, 0, sizeof(id));
  for (i = 0; i < SHA_DIGEST_LENGTH; i++) {
    (void)snprintf(id + i * 2, sizeof(id), "%02x", digest[i]);
  }

  key.data = id;
  key.size = strlen(id);
  i = asprintf((char **)&val.data, "%ld", now.tv_sec);
  if (i == -1) {
    return NULL;
  }
  val.size = strlen(val.data);

  return session->put(session, &key, &val, R_NOOVERWRITE) == 0 ?
						    strdup(id) : NULL;
}

/*
 * FUNCTION NAME:	sessionGenerateID2
 * DESCRIPTION:		generate a session id
 * ARGUMENTS:		string, string
 * RETURN VALUE:	Successful: ID, otherwise: NULL
 */
String
sessionGenerateID2(text1, text2)
	const String	text1;
	const String	text2;
{
  int			n;
  unsigned char *	text;

  n = asprintf((char **)&text, "%ld%s%u%s%u%s%ld",
		(long)time(NULL), text1, getpid(),
		envRemoteAddr(), envRemotePort(), text2,  random());
  if (n != -1) {
    String	id;
    
    id = sessionGenerateID1(text);
    free((unsigned char *)text);

    return id;
  }

  return NULL;
}

/*
 * FUNCTION NAME:	sessionValidateID
 * DESCRIPTION:		validate a session id
 * ARGUMENTS:		id
 * RETURN VALUE:	Successful: 0, Expired: 1, otherwise: -1
 */
int
sessionValidateID(id)
	const String	id;
{
  DBT	key;
  DBT	ret;

  key.data = id;
  key.size = strlen(id);
  if (session->get(session, &key, &ret, 0) == 0) {
    int	n = expireSession(&key, &ret);
    switch (n) {
      case -1:	/* error */
	return -1;
      case 0:	/* expired */
	return 1;
      default:
	return cgi_errno == CGIERR_NO_SESSION_ID ? -1: 0;
    }
  }
  cgi_errno = CGIERR_INVALID_SESSION_ID;

  return -1;
}

/*
 * FUNCTION NAME:	sessionDeleteID
 * DESCRIPTION:		delete a session id
 * ARGUMENTS:		id
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
sessionDeleteID(id)
	const String	id;
{
  DBT	key;
  DBT	ret;

  key.data = id;
  key.size = strlen(id);
  return session->del(session, &key, 0) == -1 ? -1 : 0;
}


/*
 * FUNCTION NAME:	sessionSetExpires
 * DESCRIPTION:		set a session expiry
 * ARGUMENTS:		minutes
 * RETURN VALUE:	none
 */
void
sessionSetExpires(minutes)
	unsigned int	minutes;
{
  DBT key;
  DBT val;

  if (minutes == 0) {
    minutes = DEF_SESSION_EXPIRES;
  }
  expires = minutes * 60;
  waiting = (minutes + DELETE_WAITING) * 60;

  key.data = KEY_EXPIRES;
  key.size = strlen(KEY_EXPIRES);
  val.size = asprintf(&val.data, "%u", minutes);
  if (val.size != -1) {
    session->put(session, &key, &val, 0);
  }
}

/*
 * FUNCTION NAME:	sessionGetExpires
 * DESCRIPTION:		get a session expiry
 * ARGUMENTS:		none
 * RETURN VALUE:	successful: minutes, otherwise: -1
 */
unsigned int
sessionGetExpires(void)
{
  DBT		key;
  DBT		val;

  key.data = KEY_EXPIRES;
  key.size = strlen(KEY_EXPIRES);
  if (session->get(session, &key, &val, 0) == 0) {
    String		ret;
    unsigned int	minutes;

    ret = (String)calloc(val.size + 1, sizeof(char));
    if (ret == NULL) {
      return 0;	/* error */
    }
    memcpy(ret, val.data, val.size);

    minutes = (unsigned int)atoi(ret);
    free((String)ret);

    return minutes == 0 ? DEF_SESSION_EXPIRES : minutes;
  }

  return -1;	/* error */
}

/*
 * FUNCTION NAME:	expireSession
 * DESCRIPTION:		expire the session id
 * ARGUMENTS:		key, val
 * RETURN VALUE:	Expired: 0, error: -1, otherwise: 1
 */
static int
expireSession(key, val)
	const DBT *	key;	/* id */
	const DBT *	val;	/* expired */
{
  struct timeval	now;
  String		s;
  long			date;

  if (gettimeofday(&now, NULL) == -1) {
    cgi_errno = CGIERR_NO_CURRENT_TIME;
    return -1;
  }

  s = (String)calloc(val->size + 1, sizeof(char));
  if (s == NULL) {
    cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
    return -1;
  }
  memcpy(s, val->data, val->size);
  date = atol(s);
  free((String)s);

  if (now.tv_sec - date > expires) {
    cgi_errno = CGIERR_EXPIRED_SESSION_ID;
    if (now.tv_sec - date > waiting) {
      int	n = session->del(session, key, 0);
      switch (n) {
	case 0:
	  return 0;
	case 1:
	  cgi_errno = CGIERR_NO_SESSION_ID;
	  return 1;
	default:
	  cgi_errno = CGIERR_SESSION_ID_DELETE_FAILURE;
	  return -1;
      }
    }
    return 0;
  }

  return 1;
}

static int
checkExpireSession(void)
{
  DBT	key;
  DBT	val;
  int	ret;

  for (ret = session->seq(session, &key, &val, R_FIRST); ret == 0;
       ret = session->seq(session, &key, &val, R_NEXT)) {
    if (memcmp(KEY_EXPIRES, key.data, key.size) == 0) {
      continue;
    }
    expireSession(&key, &val);
  }

  return 0;
}


/*
 * FUNCTION NAME:	sessionCountValidID
 * DESCRIPTION:		count valid session id
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: number of active id, otherwise: -1
 */
int
sessionCountValidID(void)
{
  struct timeval	now;
  DBT			key;
  DBT			val;
  int			ret;
  int			count = 0;

  if (gettimeofday(&now, NULL) == -1) {
    cgi_errno = CGIERR_NO_CURRENT_TIME;
    return -1;
  }

  for (ret = session->seq(session, &key, &val, R_FIRST); ret == 0;
       ret = session->seq(session, &key, &val, R_NEXT)) {
    String	s;
    long	date;

    s = calloc(val.size + 1, sizeof(char));
    if (s == NULL) {
      cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
      return -1;
    }
    memcpy(s, val.data, val.size);
    date = atol(s); free((String)s);
    if (now.tv_sec - date > expires) {
#if	0
      (void)session->del(session, &key, 0);
#endif
      continue;	/* expired */
    }
    count++;
  }

  return count;
}

